
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#ifndef _MOOF_SERIALIZER_HH_
#define _MOOF_SERIALIZER_HH_

/**
 * @file Serializer.hh
 * Serialize structures and types for output on a stream.
 */

#include <stdexcept>
#include <string>
#include <ostream>

#include <boost/shared_ptr.hpp>

#include <Moof/Exception.hh>


namespace Mf {


class Serializer
{
	class Impl;
	boost::shared_ptr<Impl> impl_;

public:

	/**
	 * Construction is initialization.  Use either an output stream or a string
	 * representing a path to a fill to which the serialized data will be
	 * written (replacing any previous file).
	 * @param indent If non-empty, the string's characters will be used as
	 * indentation.  Otherwise, the serialized data will not be formatted neatly
	 * but will be tightly packed.
	 */

	Serializer(const std::string& filePath, const std::string& indent = "");
	Serializer(std::ostream& output, const std::string& indent = "");

	~Serializer();

	/**
	 * Push various types of data onto the stream.
	 */

	void push(long value);
	void push(double value);
	void push(bool value);
	void push(const std::string& value);
	void push(const std::wstring& value);
	void pushNull();

	/**
	 * Push a map onto the stream.  Each map head must be matched by a map tail.
	 */

	void pushMapHead();
	void pushMapTail();

	/**
	 * Push an array onto the stream.  Each array head must be matched by an
	 * array tail.
	 */

	void pushArrayHead();
	void pushArrayTail();

	/**
	 * Write any pending bytes to the stream.  This is called automatically by
	 * the destructor of this class.
	 */

	void flush();


	/**
	 * This exception is thrown for serializer-related exceptional errors.
	 */

	struct Exception : public Mf::Exception
	{
		enum
		{
			ARCHIVE_TERMINATED		= 1024,
			KEYS_MUST_BE_STRINGS	= 1025,
			RECURSION_TOO_DEEP		= 1026,
			ALREADY_FAILED			= 1027,
		};

		explicit Exception(unsigned error) :
			Mf::Exception(error) {}

		void raise()
		{
			throw *this;
		}

		const char* what() const throw()
		{
			switch (code)
			{
				case ARCHIVE_TERMINATED:
					return "archive is already terminated";
				case KEYS_MUST_BE_STRINGS:
					return "map keys must be strings";
				case RECURSION_TOO_DEEP:
					return "generator recursion too deep";
				case ALREADY_FAILED:
					return "generator already errored out";
			}
			return Mf::Exception::what();
		}
	};
};

} // namespace Mf


#endif // _MOOF_SERIALIZER_HH_

/** vim: set ts=4 sw=4 tw=80: *************************************************/

